package minica.batch;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.StringTokenizer;

import minica.cert.CertificateJobDescription;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * 
 * @author Dominik
 * Fileformat for Batch-File:
 * - Lines beginning with # are irgnored
 * - No in-line comments
 * - All values are semicolon seperated.
 * - Leading and trailing whitespaces are stripped.
 * - Newlines seperate different sets.
 * - No Semicolons in fields.
 * # Common Name;Password;Organisational Unit;Organisation;Country;Key Algorithm;Key Length;Signature Algorithm;Valid from;Valid to; BC-Digital Sign.; BC-Key Enc.; BC-Data Enc.; BC-Key Agreement; BC-Non-Repud.; BC-Only-Enc.; BC-Only-Dec.
 * sample.tld;123;CA;TU Darmstadt;DE;RSA;2048;MD5withRSA;dd.mm.yyyy;dd.mm.yyyy;1;1;1;1;1;0;0
 */
public class BatchFileReader {
	private List<CertificateJobDescription>	jobs	= null;

	private String							inputFile;

	private CertificateJobDescription		def;

	public BatchFileReader(String inputFile, CertificateJobDescription def) {
		this.inputFile = inputFile;
		this.def = def;
	}

	public List<CertificateJobDescription> getJobs() {
		if (this.jobs == null) {
			this.parseInput();
		}
		//System.out.println("BatchFileReader.getJobs: returning " + this.jobs.size() + " jobs.");
		return this.jobs;
	}

	private void parseInput() {
		this.jobs = new ArrayList<CertificateJobDescription>();
		File input = new File(this.inputFile);
		try {
			FileReader fr = new FileReader(input);
			BufferedReader br = new BufferedReader(fr);
			String line = null;
			while ((line = br.readLine()) != null) {
				line = line.trim();
				// skip comments
				if (line.startsWith("#"))
					continue;
				//System.out.println(line);
				// parse the line
				CertificateJobDescription job = this.parseLine(line);
				if (job != null)
					this.jobs.add(job);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private CertificateJobDescription parseLine(String line) {
		StringTokenizer tok = new StringTokenizer(line, ";");
		CertificateJobDescription job = new CertificateJobDescription(this.def);
		// set reasonable defaults
		boolean[] keyUsage = new boolean[9];
		keyUsage[0] = true;
		keyUsage[1] = true;
		keyUsage[2] = true;
		keyUsage[3] = true;
		keyUsage[4] = true;
		keyUsage[5] = false;
		keyUsage[6] = false;
		keyUsage[7] = false;
		keyUsage[8] = false;
		job.setType(CertificateJobDescription.getClientCertificate());
		GregorianCalendar now = new GregorianCalendar();
		job.setValidNotBefore(now);
		job.setValidNotAfter(new GregorianCalendar(now.get(GregorianCalendar.YEAR) + 2, now
				.get(GregorianCalendar.MONTH), now.get(GregorianCalendar.DAY_OF_MONTH)));
		SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy");
		for (int i = 1; i < 18 && tok.hasMoreTokens(); i++) {
			String token = tok.nextToken().trim();
			//System.out.println("Token: " + token);
			switch (i) {
			case 1: // common-name
				//System.out.println("Setting CN to: " + token);
				job.setCommonName(token);
				break;
			case 2: // password
				//System.out.println("Setting password to: " + token);
				job.setPKCS12Password(token);
				break;
			case 3: // oragnistional unit
				//System.out.println("Setting OU to: " + token);
				job.setOrganisationalUnit(token);
				break;
			case 4: // organistation
				//System.out.println("Setting O to: " + token);
				job.setOrganisation(token);
				break;
			case 5: // contry
				//System.out.println("Setting Country name to: " + token);
				job.setCountry(token);
				break;
			case 6: // key algorithm
				//System.out.println("Setting key algorithm name to: " + token);
				job.setAlgorithm(token);
				break;
			case 7: // key length
				//System.out.println("Setting key length to: " + token);
				job.setKeyLength(Integer.parseInt(token));
				break;
			case 8: // signature algorithm
				//System.out.println("Setting signature algorithm name to: " + token);
				job.setSignatureAlgorithm(token);
				break;
			case 9: // valid from
			{
				// if there is a parsing error we set the date to now
				Date from = new Date();
				try {
					from = sdf.parse(token);
				} catch (ParseException e) {
					// ignore
				}
				GregorianCalendar gcal = new GregorianCalendar();
				gcal.setTime(from);
				job.setValidNotBefore(gcal);
			}
				break;
			case 10: // valid to
			{
				// if there is a parsing error we set the date to now
				Date to = new Date();
				GregorianCalendar gcal = new GregorianCalendar();
				try {
					to = sdf.parse(token);
					gcal = new GregorianCalendar();
					gcal.setTime(to);
				} catch (ParseException e) {
					// set to two years in the future
					gcal.set(gcal.get(GregorianCalendar.YEAR) + 2, gcal.get(GregorianCalendar.MONTH), gcal
							.get(GregorianCalendar.DATE));
				}
				job.setValidNotAfter(gcal);
			}
				break;
			case 11: // bc-digital-sign.
				if (token.equals("1")) {
					keyUsage[0] = true;
				} else {
					keyUsage[0] = false;
				}
				break;
			case 12: // bc-key enc.
				if (token.equals("1")) {
					keyUsage[2] = true;
				} else {
					keyUsage[2] = false;
				}
				break;
			case 13: // bc-data enc.
				if (token.equals("1")) {
					keyUsage[3] = true;
				} else {
					keyUsage[3] = false;
				}
				break;
			case 14: // bc-key agreement
				if (token.equals("1")) {
					keyUsage[4] = true;
				} else {
					keyUsage[4] = false;
				}
				break;
			case 15: // bc-non-repud.
				if (token.equals("1")) {
					keyUsage[1] = true;
				} else {
					keyUsage[1] = false;
				}
				break;
			case 16: // bc-only-enc.
				if (token.equals("1")) {
					keyUsage[7] = true;
				} else {
					keyUsage[7] = false;
				}
				break;
			case 17: // bc-only-dec.
				if (token.equals("1")) {
					keyUsage[8] = true;
				} else {
					keyUsage[8] = false;
				}
				break;
			default:
				break;
			}
			/*
			 * b[0] = buttonDigitalSignatureYes.getSelection();
			 * b[1] = buttonNonRepudabilityYes.getSelection();
			 * b[2] = buttonKeyEncryptionYes.getSelection();
			 * b[3] = buttonDataEncryptionYes.getSelection();
			 * b[4] = buttonKeyAgreementYes.getSelection();
			 * b[5] = false;
			 * b[6] = false;
			 * b[7] = buttonOnlyEncryptionYes.getSelection();
			 * b[8] = buttonOnlyDecryptionYes.getSelection();
			 */
		}
		job.setKeyUsage(keyUsage);
		return job;
	}
}
