package minica.cert;

import java.io.IOException;
import java.math.BigInteger;
import java.security.GeneralSecurityException;
import java.security.InvalidAlgorithmParameterException;
import java.security.KeyPair;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.cert.CertificateEncodingException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.security.auth.x500.X500Principal;

import minica.batch.BatchFileReader;
import minica.crypto.CaCertificate;
import minica.crypto.ClientCertificate;
import minica.crypto.FlexiDigestAndSignatureGenerator;
import minica.crypto.FlexiKeyGenerator;
import minica.crypto.ICaCertificate;
import minica.crypto.IClientCertificate;
import minica.crypto.KeyStoreProvider;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;

import codec.asn1.ASN1BitString;
import codec.asn1.ASN1Exception;
import codec.asn1.ASN1ObjectIdentifier;
import codec.x509.AlgorithmIdentifier;
import codec.x509.X509Certificate;
import codec.x509.X509Extension;
import codec.x509.X509TBSCertificate;
import codec.x509.extensions.AuthorityKeyIdentifierExtension;
import codec.x509.extensions.BasicConstraintsExtension;
import codec.x509.extensions.SubjectKeyIdentifierExtension;
import de.flexiprovider.ec.ecparameters.CurveParams;
import de.flexiprovider.ec.ecparameters.CurveRegistry;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * This a Job that creates the certificates.
 * @author stephan
 *
 */
public class CreateCertificateJob extends Job {

	private X509Certificate						xc;

	private X509TBSCertificate					tbs;

	private CertificateJobDescription			description;

	private FlexiDigestAndSignatureGenerator	digestGenerator;

	private FlexiKeyGenerator					keyGenerator;

	private ICaCertificate						issuerCertificate;

	private boolean								ca		= false;

	private boolean								root	= false;

	//Constructor for Certificate

	public CreateCertificateJob(CertificateJobDescription cjd) {
		super("Creating Certificate");
		description = cjd;
		try {
			if (description.getAlgorithm().contains("EC")) {
				CurveParams params = CurveRegistry.getCurveParams(description.getECCurve());
				keyGenerator = new FlexiKeyGenerator(description.getAlgorithm(), description.getKeyLength(), params);
			} else {
				keyGenerator = new FlexiKeyGenerator(description.getAlgorithm(), description.getKeyLength(), null);
			}
			digestGenerator = new FlexiDigestAndSignatureGenerator("SHA1", description.getSignatureAlgorithm());
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
		} catch (NoSuchProviderException e) {
			e.printStackTrace();
		} catch (InvalidAlgorithmParameterException e) {
			e.printStackTrace();
		}
		//		if (cjd.getCertificateType() == CertificateJobDescription.RootCertificate) {
		//			root = true;
		//		} else if (cjd.getCertificateType() == CertificateJobDescription.CACertificate) {
		//			ca = true;
		//		}
	}

	public void setIssuerDN(String isname) {
		X500Principal xp = new X500Principal(isname);
		tbs.setIssuerDN(xp);
	}

	public void setKeyUsage(boolean[] b) throws NumberFormatException, Exception {
		ASN1BitString usage = new ASN1BitString(b);
		X509Extension keyusage = new X509Extension(new ASN1ObjectIdentifier("2.5.29.15"), true, usage);
		tbs.addExtension(keyusage);
	}

	public void setSignatureAlgorithm(String siga) throws NoSuchAlgorithmException {
		AlgorithmIdentifier ai = new AlgorithmIdentifier(siga);
		tbs.setSignatureAlgorithm(ai);
	}

	//	public void setUniqueIDs() throws NoSuchAlgorithmException, NoSuchProviderException, CertificateEncodingException, FileNotFoundException {
	//		xc.setSubjectUniqueID(digestGenerator.digest(subject));
	//		if (root) {
	//			tbs.setIssuerUniqueID(digestGenerator.digest(issuer));
	//		} else if (ca) {
	//			X509Cert issuerCertificate = new X509Cert(new DERDecoder(new FileInputStream(Application.getHome() + "root.der")));
	//			issuer = issuerCertificate.getIssuerDN().toString();
	//			tbs.setIssuerUniqueID(digestGenerator.digest(issuer));
	//		}
	//	}

	public void setSubjectDN(String isname) {
		X500Principal xp = new X500Principal(isname);
		tbs.setSubjectDN(xp);
	}

	/*
	public void writeToFile(PFX pfx, X509Certificate xc) throws IOException, ASN1Exception,
			CertificateEncodingException {

		FileOutputStream output = new FileOutputStream(new File(Application.getHome() + description.getFileName()
				+ ".der"));
		output.write(xc.getEncoded());
		output.close();

		output = new FileOutputStream(new File(Application.getHome() + description.getFileName() + ".txt"));
		output.write((xc.toString()).getBytes());
		output.close();

		DEREncoder enc = new DEREncoder(
				new FileOutputStream(Application.getHome() + description.getFileName() + ".p12"));
		pfx.encode(enc);
		enc.close();
	}
	*/

	@Override
	protected IStatus run(IProgressMonitor monitor) {
		// this was added to enable batch processing
		List<CertificateJobDescription> jobs;
		if (description.getType() == CertificateJobDescription.BatchMode) {
			// parse the input file
			BatchFileReader bfr = new BatchFileReader(description.getFileName(), description);
			jobs = bfr.getJobs();

		} else if (description.getType() == CertificateJobDescription.OneClick) {
			jobs = new ArrayList<CertificateJobDescription>();

			boolean[] userKeyUsage = new boolean[9];
			userKeyUsage[0] = true;
			userKeyUsage[1] = true;
			userKeyUsage[2] = true;
			userKeyUsage[3] = true;
			userKeyUsage[4] = true;
			userKeyUsage[5] = false;
			userKeyUsage[6] = false;
			userKeyUsage[7] = false;
			userKeyUsage[8] = false;
			description.setKeyUsage(userKeyUsage);

			CertificateJobDescription rootDescription = new CertificateJobDescription(description);

			rootDescription.setCommonName("OneClickRoot " + description.getCommonName());

			boolean[] rootKeyUsage = new boolean[9];
			rootKeyUsage[0] = false;
			rootKeyUsage[1] = false;
			rootKeyUsage[2] = false;
			rootKeyUsage[3] = false;
			rootKeyUsage[4] = false;
			rootKeyUsage[5] = true;
			rootKeyUsage[6] = false;
			rootKeyUsage[7] = false;
			rootKeyUsage[8] = false;
			rootDescription.setKeyUsage(rootKeyUsage);

			rootDescription.setPathLength(1);
			BigInteger serialOCRoot = KeyStoreProvider.getInstance().getMaxSerial().add(new BigInteger("1"));
			description.setIssuer(serialOCRoot.toString());

			rootDescription.setType(CertificateJobDescription.RootCertificate);
			description.setType(CertificateJobDescription.ClientCertificate);
			description.setExportAlternates(true);

			jobs.add(rootDescription);
			jobs.add(description);

		} else {
			jobs = new ArrayList<CertificateJobDescription>();
			jobs.add(description);
		}

		// always asume a list of jobs. if this is called in single-cert mode there will be
		// only one certificate in the list.
		monitor.beginTask("Creating Certificates ...", jobs.size());

		for (Iterator<CertificateJobDescription> iterator = jobs.iterator(); iterator.hasNext();) {
			try {
				CertificateJobDescription certificateJobDescription = (CertificateJobDescription) iterator.next();

				if (certificateJobDescription.getCertificateType() == CertificateJobDescription.RootCertificate) {
					root = true;
					ca = false;
				} else if (certificateJobDescription.getCertificateType() == CertificateJobDescription.CACertificate) {
					ca = true;
					root = false;
				} else {
					root = false;
					ca = false;
				}

				// set a unique serial for this certificate
				certificateJobDescription.setSerial(KeyStoreProvider.getInstance().getNextSerial());
				xc = new X509Certificate();
				tbs = new X509TBSCertificate();

				//Fill V1 Fields

				String subject = "CN=" + certificateJobDescription.getCommonName() + ", OU="
						+ certificateJobDescription.getOrganisationalUnit() + ", O="
						+ certificateJobDescription.getOrganisation() + ", C=" + certificateJobDescription.getCountry();
				setSubjectDN(subject);

				String issuer;

				if (root) {
					issuer = subject;
				} else {

					issuerCertificate = KeyStoreProvider.getInstance().getCertificate(
							new BigInteger(certificateJobDescription.getIssuer()));
					issuer = issuerCertificate.getCertificate().getIssuerDN().toString();
				}

				setIssuerDN(issuer);

				tbs.setNotBefore(certificateJobDescription.getValidNotBefore());
				tbs.setNotAfter(certificateJobDescription.getValidNotAfter());
				tbs.setSerialNumber(certificateJobDescription.getSerial());

				setSignatureAlgorithm(certificateJobDescription.getSignatureAlgorithm());
				//Unique Identifiers
				//setUniqueIDs();

				//Key Usage
				setKeyUsage(certificateJobDescription.getKeyUsage());

				//Create Keypair
				KeyPair kp = keyGenerator.createKey();
				tbs.setSubjectPublicKey(kp.getPublic());

				//SKIE
				SubjectKeyIdentifierExtension skie = new SubjectKeyIdentifierExtension();
				skie.setKeyIdentifier(digestGenerator.digest(kp.getPublic().getEncoded()));
				tbs.addExtension(skie);

				//AKIE
				AuthorityKeyIdentifierExtension akie = new AuthorityKeyIdentifierExtension();
				if (root) {
					akie.setCertSerialNumber(certificateJobDescription.getSerial());
					akie.setKeyIdentifier(digestGenerator.digest(kp.getPublic().getEncoded()));
				} else {

					akie.setCertSerialNumber(issuerCertificate.getCertificate().getSerialNumber());
					akie.setKeyIdentifier(digestGenerator.digest(issuerCertificate.getCertificate().getPublicKey()
							.getEncoded()));
				}

				tbs.addExtension(akie);

				//Basic Constraints
				if (root | ca) {
					BasicConstraintsExtension bc = new BasicConstraintsExtension(true, certificateJobDescription
							.getPathlength());
					bc.setExplicit(true);
					tbs.addExtension(bc);
				}

				//System.out.println(tbs.getSigAlgName());

				xc = new X509Certificate(tbs);

				//Sign
				if (root) {
					digestGenerator.sign(kp.getPrivate(), kp.getPublic(), xc);
					CaCertificate cacert = new CaCertificate(xc, null);
					cacert.setKeyPair(kp, certificateJobDescription.getPkcs12Password());
					// we need this addCert call because the parent cert is null
					KeyStoreProvider.getInstance().addCert(cacert);
				} else if (ca) {
					PrivateKey pk = issuerCertificate.getKeyPair(certificateJobDescription.getIssuerPassword())
							.getPrivate();
					digestGenerator.sign(pk, kp.getPublic(), xc);
					ICaCertificate cacert = new CaCertificate(xc, issuerCertificate);
					cacert.setKeyPair(kp, certificateJobDescription.getPkcs12Password());
					// no KeyStoreProvider.getInstace().addCert necessary. The constructor of CaCerficiate handles everything
					// as long as the parent cert is != null.
				} else {
					PrivateKey pk = issuerCertificate.getKeyPair(certificateJobDescription.getIssuerPassword())
							.getPrivate();
					digestGenerator.sign(pk, kp.getPublic(), xc);
					// no KeyStoreProvider.getInstace().addCert necessary. The constructor of ClientCerficiate handles everything
					// as long as the parent cert is != null.
					IClientCertificate clientCert = new ClientCertificate(xc, issuerCertificate);
					if (certificateJobDescription.isExportAlternates()) {
						clientCert.setKeyPair(kp, certificateJobDescription.getPkcs12Password(), true);
					} else {
						clientCert.setKeyPair(kp, certificateJobDescription.getPkcs12Password());
					}
				}

			} catch (NoSuchAlgorithmException e) {
				e.printStackTrace();
			} catch (CertificateEncodingException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (ASN1Exception e) {
				e.printStackTrace();
			} catch (GeneralSecurityException e) {
				e.printStackTrace();
			} catch (Exception e) {
				e.printStackTrace();
			}
			monitor.worked(1);
		}
		// save the data-structure
		KeyStoreProvider.getInstance().serialize();
		monitor.done();
		return Status.OK_STATUS;
	}
}
