package minica.crl;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.PublicKey;
import java.security.cert.CRLException;
import java.util.ArrayList;
import java.util.GregorianCalendar;

import minica.Application;
import minica.crypto.FlexiDigestAndSignatureGenerator;
import minica.crypto.ICaCertificate;
import minica.util.CRLHelper;
import minica.wizards.CRLJobDescription;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;

import codec.asn1.ASN1Integer;
import codec.asn1.ASN1ObjectIdentifier;
import codec.x501.BadNameException;
import codec.x501.Name;
import codec.x509.AlgorithmIdentifier;
import codec.x509.CRLEntry;
import codec.x509.X509Crl;
import codec.x509.X509Extension;
import codec.x509.extensions.AuthorityKeyIdentifierExtension;
import codec.x509.extensions.ReasonCodeExtension;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * This is a Job which can create a CRL or Delta-CRL.
 * 
 * @author stephan
 *
 */
public class CreateCRLJob extends Job {

	private CRLJobDescription	description;

	private ICaCertificate		issuer;

	private X509Crl				crl;

	private boolean				createDeltaCRL	= false;

	/**
	 * Create a CRL Job.
	 * @param description The JobDescription with the work to be done
	 */

	public CreateCRLJob(CRLJobDescription description) {
		super("Creating CRL");
		this.description = description;
		this.issuer = description.getIssuer();
	}

	public CreateCRLJob(String name) {
		super(name);
	}

	/**
	 * if the argument is true, a Delta CRL will be created.
	 * @param delta
	 */

	public void setDeltaCRL(boolean delta) {
		this.createDeltaCRL = delta;
	}

	/**
	 * Write the new CRL to the Harddrive. 
	 * @param filename the filename without extension.
	 */

	public void writeToFile(String filename) {
		FileOutputStream output;
		try {
			output = new FileOutputStream(new File(Application.getHome() + description.getFileName() + ".crl"));

			output.write(crl.getEncoded());
			output.close();

		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (CRLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	@SuppressWarnings("unchecked")
	@Override
	//Code to create the CRL
	protected IStatus run(IProgressMonitor monitor) {
		try {
			ArrayList<Integer> reasoncodes = description.getReasonCodes();
			ArrayList<BigInteger> toBeRevokedCertificates = description.getToBeRevokedCertificates();

			String tempname = description.getIssuerDN().toString();
			Name issuername = new Name(tempname, -1);

			GregorianCalendar cal = new GregorianCalendar();

			crl = new X509Crl(issuername, cal);
			X509Crl oldcrl = issuer.getCRL();
			CRLHelper helper = new CRLHelper(oldcrl);

			if (description.getType() == CRLJobDescription.DELTACRL) {
				createDeltaCRL = true;
			}

			//Set CRL Number, Add revoked Certificates from last CRL

			if (oldcrl == null) {
				X509Extension CRLNumber = new X509Extension(new ASN1ObjectIdentifier("2.5.29.20"), false,
						new ASN1Integer(1));
				crl.addExtension(CRLNumber);

			} else {

				ArrayList<CRLEntry> revokedCerts = new ArrayList<CRLEntry>(oldcrl.getRevokedCertificates());

				if (!createDeltaCRL) {

					for (int i = 0; i < revokedCerts.size(); i++) {
						crl.addEntry(revokedCerts.get(i));
					}
				}

				//If a deltaCRL was issued 

				if (issuer.getDeltaCrl() != null) {

					X509Crl deltacrl = issuer.getDeltaCrl();

					//Add revoked Certificates from last Delta CRL

					ArrayList<CRLEntry> revokedCertsDelta = new ArrayList<CRLEntry>(deltacrl.getRevokedCertificates());
					for (int i = 0; i < revokedCertsDelta.size(); i++) {
						crl.addEntry(revokedCertsDelta.get(i));
					}

					//Set the CRL Serial Number to deltacrlSerialNumber + 1

					CRLHelper helperDelta = new CRLHelper(deltacrl);

					X509Extension CRLNumber = new X509Extension(new ASN1ObjectIdentifier("2.5.29.20"), false,
							new ASN1Integer(helperDelta.getSerial().add(new BigInteger("1"))));
					crl.addExtension(CRLNumber);

				} else {

					//If no DeltaCRL was present, set serial to lastcrl serial + 1

					X509Extension CRLNumber = new X509Extension(new ASN1ObjectIdentifier("2.5.29.20"), false,
							new ASN1Integer(helper.getSerial().add(new BigInteger("1"))));
					crl.addExtension(CRLNumber);
				}
			}

			// Delta CRL Indicator
			if (createDeltaCRL) {
				X509Extension deltaCRLIndicator = new X509Extension(new ASN1ObjectIdentifier("2.5.29.27"), true,
						new ASN1Integer(helper.getSerial()));
				crl.addExtension(deltaCRLIndicator);

			}

			//Create fresh revoked Entries

			for (int i = 0; i < reasoncodes.size(); i++) {
				CRLEntry entry = new CRLEntry(toBeRevokedCertificates.get(i), cal);
				ReasonCodeExtension reason = new ReasonCodeExtension(reasoncodes.get(i));
				entry.addExtension(reason);
				crl.addEntry(entry);
			}

			// Update Date
			crl.setThisUpdate(cal);
			crl.setNextUpdate(description.getNextUpdate());

			//AKIE

			PublicKey pk = issuer.getCertificate().getPublicKey();

			FlexiDigestAndSignatureGenerator signatureGenerator = new FlexiDigestAndSignatureGenerator("SHA1", issuer
					.getCertificate().getSigAlgName());

			AuthorityKeyIdentifierExtension akie = new AuthorityKeyIdentifierExtension();

			akie.setKeyIdentifier(signatureGenerator.digest(pk.getEncoded()));
			akie.setCertSerialNumber(description.getIssuer().getCertificate().getSerialNumber());

			crl.addExtension(akie);

			//Algorithm Identifier

			AlgorithmIdentifier ai = new AlgorithmIdentifier(issuer.getCertificate().getSigAlgName());
			crl.setSignatureAlgorithm(ai);

			//Create Signature

			byte[] signature = signatureGenerator.getSignature(description.getIssuer().getKeyPair(
					description.getIssuerPassword()).getPrivate(), crl.getTBSCertList(ai));
			crl.setSignature(signature);

			if (createDeltaCRL) {
				issuer.setDeltaCrl(crl);
			} else {
				issuer.setCRL(crl);
			}

			writeToFile(description.getFileName());

		} catch (BadNameException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return Status.OK_STATUS;
	}

}
