package minica.crypto;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateEncodingException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import minica.Application;
import codec.asn1.ASN1Exception;
import codec.asn1.DERDecoder;
import codec.pkcs12.PFX;
import codec.pkcs12.PKCS8ShroudedKeyBag;
import codec.pkcs12.SafeBag;
import codec.x509.X509Certificate;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * An abstract certificate.
 * @author dominik
 */
public class AbstractCertificate implements ChangeListener {

	private PublicKey				pk;

	private List<ChangeListener>	listener	= new ArrayList<ChangeListener>();

	protected X509Certificate		cert;

	protected ICaCertificate		parentCert;

	protected String				pkcs12Filename;

	public AbstractCertificate() {
		super();
	}

	public void addChangeListener(ChangeListener l) {
		this.listener.add(l);

	}

	public X509Certificate getCertificate() {
		return this.cert;
	}

	public KeyPair getKeyPair(String password) throws GeneralSecurityException {
		// read the private key from the PKCS12
		PrivateKey privk = null;
		try {
			privk = this.readPrivateKeyFromPKCS12(this.pkcs12Filename, password);
		} catch (IllegalStateException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ASN1Exception e) {
			e.printStackTrace();
		}
		return new KeyPair(this.pk, privk);
	}

	public PrivateKey readPrivateKeyFromPKCS12(String filename, String password) throws GeneralSecurityException,
			IllegalStateException, IOException, ASN1Exception {
		DERDecoder decoder = new DERDecoder(new FileInputStream(Application.getHome() + filename + "."
				+ KeyStoreProvider.PKCS12_EXTENSION));
		PFX pfx = new PFX();
		pfx.decode(decoder);
		SafeBag safeBag = pfx.getAuthSafe().getSafeContents(0).getSafeBag(0);
		PKCS8ShroudedKeyBag shroudedKeyBag = (PKCS8ShroudedKeyBag) safeBag.getBagValue();
		return shroudedKeyBag.getPrivateKey(password.toCharArray());
	}

	public void removeChangeListener(ChangeListener l) {
		this.listener.remove(l);
	}

	public void setKeyPair(KeyPair kp, String password) {
		this.setKeyPair(kp, password, false);
	}

	public void setKeyPair(KeyPair kp, String password, boolean exportAlternates) {
		// write the private key to the PKCS12
		this.pk = kp.getPublic();
		File pkFile = new File(Application.getHome() + this.pkcs12Filename);
		try {
			if (exportAlternates) {
				Export.writeToPkcs12File(kp, this.cert, password, pkFile.getAbsolutePath(), true, true, true, true,
						true);
			} else {
				Export.writeToPkcs12File(kp, this.cert, password, pkFile.getAbsolutePath());
			}
		} catch (CertificateEncodingException e) {
			e.printStackTrace();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (ASN1Exception e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void stateChanged(ChangeEvent e) {
		for (ChangeListener l : this.listener) {
			l.stateChanged(e);
		}
	}
}
