package minica.crypto;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import javax.swing.event.ChangeEvent;

import codec.x509.X509Certificate;
import codec.x509.X509Crl;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * A CA Certificate, either root ca or sub-ca.
 * @author tex
 *
 */
public class CaCertificate extends AbstractCertificate implements ICaCertificate {

	/**
	 * Save this CA's most recent CRL here.
	 */
	private X509Crl						crl;

	/**
	 * Save this CA's delta CRL here.
	 */
	private X509Crl						deltaCrl;

	private List<IClientCertificate>	clientCerts;

	private List<ICaCertificate>		subCerts;

	public CaCertificate(X509Certificate cert, ICaCertificate parent) {
		this.cert = cert;
		this.parentCert = parent;
		this.clientCerts = new ArrayList<IClientCertificate>();
		this.subCerts = new ArrayList<ICaCertificate>();
		/*
		 * Important: If we have a parent we'll add ourselves to
		 * it's list of certificates. The parent in turn will
		 * register itself in our list of listeners so we can tell him
		 * if anything happens to us (new sub-certs, etc.).
		 * Nice, 'eh? ;D 
		 */
		if (this.parentCert != null) {
			this.parentCert.addCert(this);
		}
		if (this.cert != null) {
			this.pkcs12Filename = this.cert.getSerialNumber().toString();
		}
		this.stateChanged(new ChangeEvent(this));
	}

	/**
	 * Recursively (try to) add a certificate.
	 */
	public boolean addCert(ICertificate cert) {
		if (cert != null && cert.getParent() != null) {
			ICertificate parent = cert.getParent();
			if (this.equals(parent)) {
				cert.addChangeListener(this);
				if (cert instanceof IClientCertificate) {
					IClientCertificate client = (IClientCertificate) cert;
					if (!this.clientCerts.contains(client)) {
						//System.out.println("CaCertificate.addCert - Adding client cert: " + client.getCertificate().getSerialNumber());
						this.clientCerts.add(client);
						this.stateChanged(new ChangeEvent(this));
						/*
						System.out.println("CaCertificate(" + this.cert.getSerialNumber()
								+ ").addCert: Dumping a list of conained certificates: ");
						for (int i = 0; i < this.getCertificates().size(); i++) {
							ICertificate c = this.getCertificates().get(i);
							if (c instanceof ICaCertificate)
								System.out.println("CA: " + this.getCertificates().get(i).getCertificate().getSerialNumber());
							else if (c instanceof IClientCertificate)
								System.out.println("Client: " + this.getCertificates().get(i).getCertificate().getSerialNumber());
						}
						*/
						return true;
					}
				} else if (cert instanceof ICaCertificate) {
					ICaCertificate ca = (ICaCertificate) cert;
					if (!this.subCerts.contains(ca)) {
						this.subCerts.add(ca);
						this.stateChanged(new ChangeEvent(this));
						return true;
					}
				}
			}
		}
		return false;
	}

	public boolean contains(Object o) {
		if (this.equals(o)) {
			return true;
		}
		// check the client certs
		for (int i = 0; i < this.clientCerts.size(); i++) {
			if (this.clientCerts.get(i).equals(o)) {
				return true;
			}
		}
		// check the sub certs
		for (int i = 0; i < this.subCerts.size(); i++) {
			// check if o is a direct child of this cert.
			if (this.subCerts.get(i).equals(o)) {
				return true;
			}
			// check if o is some grand child of this cert
			// by asking our childs
			if (this.subCerts.get(i).contains(o)) {
				return true;
			}
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final CaCertificate other = (CaCertificate) obj;
		if (cert == null) {
			if (other.cert != null)
				return false;
		} else if (!cert.equals(other.cert))
			return false;
		if (clientCerts == null) {
			if (other.clientCerts != null)
				return false;
		} else if (!clientCerts.equals(other.clientCerts))
			return false;
		if (crl == null) {
			if (other.crl != null)
				return false;
		} else if (!crl.equals(other.crl))
			return false;
		if (subCerts == null) {
			if (other.subCerts != null)
				return false;
		} else if (!subCerts.equals(other.subCerts))
			return false;
		return true;
	}

	public ICaCertificate get(Object o) {
		if (this.equals(o)) {
			return this;
		}
		for (int i = 0; i < this.subCerts.size(); i++) {
			if (this.subCerts.get(i).equals(o)) {
				return this.subCerts.get(i);
			}
			if (this.subCerts.get(i).contains(o)) {
				return this.subCerts.get(i).get(o);
			}
		}
		return null;
	}

	public List<ICaCertificate> getAllCaCerts() {
		List<ICaCertificate> l = new ArrayList<ICaCertificate>();
		for (int i = 0; i < this.subCerts.size(); i++) {
			l.add(this.subCerts.get(i));
			l.addAll(this.subCerts.get(i).getAllCaCerts());
		}
		return l;
	}

	public List<ICaCertificate> getCaCerts() {
		return this.subCerts;
	}

	public List<ICertificate> getCertificates() {
		List<ICertificate> c = new ArrayList<ICertificate>();
		c.addAll(this.subCerts);
		c.addAll(this.clientCerts);
		//System.out.println("CaCertificate.getCertificates("+this.cert.getSerialNumber()+") - num certs: " + c.size());
		return c;
	}

	public IClientCertificate[] getClientCertificates() {
		IClientCertificate[] a = new IClientCertificate[1];
		return this.clientCerts.toArray(a);
	}

	public List<IClientCertificate> getClientCerts() {
		return this.clientCerts;
	}

	public X509Crl getCRL() {
		return this.crl;
	}

	/**
	 * @return the deltaCrls
	 */
	public X509Crl getDeltaCrl() {
		return deltaCrl;
	}

	public ICertificate getParent() {
		return this.getParentCa();
	}

	public ICaCertificate getParentCa() {
		return this.parentCert;
	}

	public ICertificate getSubCert(BigInteger serial) {
		for (ICaCertificate cert : subCerts) {
			if (cert.getCertificate().getSerialNumber().equals(serial)) {
				return cert;
			}
		}
		for (ICertificate cert : clientCerts) {
			if (cert.getCertificate().getSerialNumber().equals(serial)) {
				return cert;
			}
		}
		return null;
	}

	public ICaCertificate[] getSubCertificates() {
		ICaCertificate[] a = new ICaCertificate[1];
		return this.subCerts.toArray(a);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((cert == null) ? 0 : cert.hashCode());
		result = prime * result + ((clientCerts == null) ? 0 : clientCerts.hashCode());
		result = prime * result + ((crl == null) ? 0 : crl.hashCode());
		result = prime * result + ((subCerts == null) ? 0 : subCerts.hashCode());
		return result;
	}

	/**
	 * Recursively (try to) remove a certificate.
	 */
	public boolean removeCert(ICertificate cert) {
		if (cert != null && cert.getParent() != null && this.equals(cert.getParent())) {
			if (cert instanceof ICaCertificate) {
				this.subCerts.remove((ICaCertificate) cert);
				this.stateChanged(new ChangeEvent(this));
				return true;
			} else if (cert instanceof IClientCertificate) {
				this.clientCerts.remove((IClientCertificate) cert);
				this.stateChanged(new ChangeEvent(this));
				return true;
			} else {
				return false;
			}
		} else {
			for (int i = 0; i < this.subCerts.size(); i++) {
				if (this.subCerts.get(i).removeCert(cert)) {
					this.stateChanged(new ChangeEvent(this));
					return true;
				}
			}
			return false;
		}
	}

	public void setCertificate(X509Certificate cert) {
		this.cert = cert;
	}

	public void setCRL(X509Crl crl) {
		this.crl = crl;
	}

	/**
	 * @param deltaCrl the deltaCrl to set
	 */
	public void setDeltaCrl(X509Crl deltaCrl) {
		this.deltaCrl = deltaCrl;
	}

	public void setParentCa(ICaCertificate parent) {
		this.parentCert = parent;
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("CaCertificate[");
		sb.append("Cert: ");
		if (this.cert != null)
			sb.append(this.cert.getSerialNumber() + ", " + this.cert.getSubjectDN());
		sb.append(", CRL: ");
		if (this.crl != null)
			sb.append(this.crl.toString());
		sb.append(", ClientCerts: ");
		if (this.clientCerts != null) {
			for (int i = 0; i < this.clientCerts.size(); i++) {
				sb.append(this.clientCerts.get(i).toString() + ", ");
			}
		}
		sb.append(", SubCerts: ");
		if (this.subCerts != null) {
			for (int i = 0; i < this.subCerts.size(); i++) {
				sb.append(this.subCerts.get(i).toString() + ", ");
			}
		}
		sb.append("]");
		return sb.toString();
	}

}
