package minica.crypto;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.Provider;
import java.security.Security;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import minica.Application;
import minica.util.XMLHelper;

import org.eclipse.core.runtime.jobs.Job;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import de.flexiprovider.core.FlexiCoreProvider;
import de.flexiprovider.ec.FlexiECProvider;
import de.flexiprovider.nf.FlexiNFProvider;
import de.flexiprovider.pqc.FlexiPQCProvider;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * The main datastructure that contains the information about the available algorithms.
 * Uses the singleton pattern.
 * @author tex
 *
 */
public class CryptoInformationProvider {

	private static final String					MINICA_ALGODB_XML	= "minica_algodb.xml";

	private static final String					ALGODB_STRENGTH		= "Strength";

	private static final String					ALGODB_ASYMMETRIC	= "Asymmetric";

	private static final String					ALGODB_SYMMETRIC	= "Symmetric";

	private static final String					ALGODB_SIGNATURE	= "Signature";

	private static final String					ALGODB_DIGEST		= "Digest";

	private static final String					ALGODB_PROVIDER		= "Provider";

	private static final String					ALGODB_ROOTNODE		= "Minica";

	/**
	 * The instance of the class. Part of the singleton pattern.
	 */
	private static CryptoInformationProvider	instance;

	/**
	 * The getter for the instance of this class. Part of
	 * the singleton pattern. Needs to be syncronized or
	 * else there will be more instances of this class due to race-conditions.
	 * @return
	 */
	public synchronized static CryptoInformationProvider getInstance() {
		if (null == instance) {
			instance = new CryptoInformationProvider();
		}
		return instance;
	}

	/**
	 * Store a list of all working symmetric algorithms and their available keysizes.
	 */
	private Map<String, Set<Integer>>	symmetricStrenghts	= new HashMap<String, Set<Integer>>();

	/**
	 * Store a list of all working asymmetric algorithms and their available keysizes.
	 */
	private Map<String, Set<Integer>>	asymmetricStrenghts	= new HashMap<String, Set<Integer>>();

	/**
	 * Store a list of all working message digest algorithms.
	 */
	private Set<String>					digestAlgorithm		= new HashSet<String>();

	/**
	 * Store a list of all working signature algorithms.
	 */
	private Set<String>					signatureAlgorithms	= new HashSet<String>();

	private CryptoInformationProvider() {
		Security.addProvider(new FlexiCoreProvider());
		Security.addProvider(new FlexiECProvider());
		Security.addProvider(new FlexiNFProvider());
		Security.addProvider(new FlexiPQCProvider());
		//Security.addProvider(new SunJCE());
		File algodbFile = new File(Application.getHome() + MINICA_ALGODB_XML);
		try {
			//System.out.println("Cip() - Trying to read the algodb at " + algodbFile.getAbsolutePath());
			this.readDb(algodbFile);
		} catch (IllegalArgumentException e) {
			//System.out.println("Cip() - reading algodb failed. Exception: " + e);
			this.updateDb();
		}
	}

	/**
	 * Add a known-valid asymmetric algorithm-keysize combination
	 * NOTE: Not (yet) thread-safe!
	 * @param name the algorithm name
	 * @param keysize a working keysize
	 */
	public synchronized void addValidAsymmKeySize(String name, int keysize) {
		if (this.asymmetricStrenghts.containsKey(name)) {
			this.asymmetricStrenghts.get(name).add(new Integer(keysize));
		} else {
			Set<Integer> s = new HashSet<Integer>();
			s.add(new Integer(keysize));
			this.asymmetricStrenghts.put(name, s);
		}
	}

	/**
	 * Add a known-valid digest algorithm.
	 * NOTE: Not (yet) thread-safe!
	 * @param name
	 */
	public synchronized void addValidDigestAlgorithm(String name) {
		this.digestAlgorithm.add(name);
	}

	/**
	 * Add a known-valid signature algorithm
	 * NOTE: Not (yet) thread-safe!
	 * @param name 
	 */
	public synchronized void addValidSignatureAlgorithm(String name) {
		this.signatureAlgorithms.add(name);
	}

	/**
	 * Adds a known-valid symmetric algorithm-keysize combination.
	 * NOTE: Not (yet) thread-safe!
	 * @param name
	 * @param keysize
	 */
	public synchronized void addValidSymmKeySize(String name, int keysize) {
		if (this.symmetricStrenghts.containsKey(name)) {
			this.symmetricStrenghts.get(name).add(new Integer(keysize));
		} else {
			Set<Integer> s = new HashSet<Integer>();
			s.add(new Integer(keysize));
			this.symmetricStrenghts.put(name, s);
		}
	}

	/**
	 * 
	 * @return all available asymmetric algorithms
	 */
	public Map<String, Set<Integer>> getValidAsymmetricAlgorithms() {
		return this.asymmetricStrenghts;
	}

	/**
	 * @return all available digest algorithms
	 */
	public Set<String> getValidDigestAlgorithms() {
		return this.digestAlgorithm;
	}

	/**
	 * Get all valid key sizes for the given algorithm.
	 * returns 0 if name is not found.
	 * @param name
	 * @return
	 */
	public Set<Integer> getValidKeySizesForAsym(String name) {
		if (this.asymmetricStrenghts.containsKey(name)) {
			return this.asymmetricStrenghts.get(name);
		} else {
			Set<Integer> s = new HashSet<Integer>();
			s.add(new Integer(0));
			return s;
		}
	}

	/**
	 * Get all valid key sizes for the given algorithm.
	 * retruns 0 if name is not found.
	 * @param name
	 * @return
	 */
	public Set<Integer> getValidKeySizesForSymm(String name) {
		if (this.symmetricStrenghts.containsKey(name)) {
			return this.symmetricStrenghts.get(name);
		} else {
			Set<Integer> s = new HashSet<Integer>();
			s.add(new Integer(0));
			return s;
		}
	}

	/**
	 * 
	 * @return all available signature algorithms
	 */
	public Set<String> getValidSignatureAlgorithms() {
		return this.signatureAlgorithms;
	}

	public Set<String> getValidSignatureAlgorithmsFor(String keyAlgorithm) {

		Set<String> all = this.getValidSignatureAlgorithms();
		String[] avail = Application.getSettings().getArray("WorkingSignatures");
		Set<String> apt = new HashSet<String>();
		//		for (Iterator iterator = avail.iterator(); iterator.hasNext();) {
		//			String alg = (String) iterator.next();
		//			if (alg.contains("with" + keyAlgorithm) && !alg.contains("NONE")) {
		//				apt.add(alg);
		//			}
		//		}

		for (int i = 0; i < avail.length; i++) {
			String alg = avail[i];
			String edited = alg.substring(0, alg.indexOf(","));
			if (alg.contains("with" + keyAlgorithm) && !alg.contains("NONE")) {
				apt.add(edited);
			}
		}

		// if no appropriate signature algorithms were found,
		// return all an let the user decide.
		if (apt.size() < 1) {
			return all;
		} else {
			return apt;
		}
	}

	/**
	 * Returns all valid symmetric algorithms.
	 * @return
	 */
	public Map<String, Set<Integer>> getValidSymmetricAlgorithms() {
		return this.symmetricStrenghts;
	}

	public void saveDb() {
		File dbFile = new File(Application.getHome() + CryptoInformationProvider.MINICA_ALGODB_XML);
		this.saveDb(dbFile);
	}

	/**
	 * Read the stored algorithm db.
	 * Should be called from the constructor.
	 * @param xmlFile
	 */
	@SuppressWarnings("unchecked")
	private void readDb(File xmlFile) {
		// deserializeation of the data
		if (!xmlFile.exists() || !xmlFile.canRead()) {
			throw new IllegalArgumentException("File not accessible! at Location " + xmlFile.getAbsolutePath());
		}
		boolean dbValid = true;
		SAXBuilder xmlBuilder = new SAXBuilder();
		Document doc;
		try {
			doc = xmlBuilder.build(new FileInputStream(xmlFile));
			Element root = doc.getRootElement();
			List rootElements = root.getChildren();
			// Read the provider and check for validity (hashCode)
			Element providers = XMLHelper.getXMLNodeByName(rootElements, ALGODB_PROVIDER);
			if (null == providers) {
				// need to redo the algodb
				dbValid = false;
			}
			// provider auslesen und vergleichen
			List providerList = providers.getChildren();
			for (int i = 0; i < providerList.size(); i++) {
				Element provider = (Element) providerList.get(i);
				if (null == provider) {
					continue;
				}
				int hashsum = Integer.parseInt(provider.getText());
				Provider prov = Security.getProvider(provider.getName());
				// check if the class is the same
				if (hashsum != prov.hashCode()) {
					dbValid = false;
				}
			}
			// Read the digest algorithms
			Element digests = XMLHelper.getXMLNodeByName(rootElements, ALGODB_DIGEST);
			if (null == digests) {
				// FIXME need to redo the digest part of the db
			} else {
				List digestList = digests.getChildren();
				for (int i = 0; i < digestList.size(); i++) {
					Element child = (Element) digestList.get(i);
					if (null == child) {
						continue;
					}
					this.digestAlgorithm.add(child.getName());
				}
			}
			// Read the signature algorithms
			Element signatures = XMLHelper.getXMLNodeByName(rootElements, ALGODB_SIGNATURE);
			if (null == signatures) {
				// FIXME need to redo the sign. part of the db
			} else {
				List signList = signatures.getChildren();
				for (int i = 0; i < signList.size(); i++) {
					Element child = (Element) signList.get(i);
					if (null == child) {
						continue;
					}
					this.signatureAlgorithms.add(child.getName());
				}
			}
			// Read the symmetric algorithms
			Element symmetrics = XMLHelper.getXMLNodeByName(rootElements, ALGODB_SYMMETRIC);
			if (null == symmetrics) {
				// FIXME need to redo the symmetric part of the db
			} else {
				List symmetricList = symmetrics.getChildren();
				for (int i = 0; i < symmetricList.size(); i++) {
					Element child = (Element) symmetricList.get(i);
					if (null == child) {
						continue;
					}
					String algoName = child.getName();
					Set<Integer> keysizes = new HashSet<Integer>();
					List keysizesList = child.getChildren();
					for (int j = 0; j < keysizesList.size(); j++) {
						Element keySize = (Element) keysizesList.get(j);
						if (keySize == null) {
							continue;
						}
						Integer ks = Integer.parseInt(keySize.getText());
						keysizes.add(ks);
					}
					this.symmetricStrenghts.put(algoName, keysizes);
				}
			}
			// read the asymmetric algorithms
			Element asymmetrics = XMLHelper.getXMLNodeByName(rootElements, ALGODB_ASYMMETRIC);
			if (null == asymmetrics) {
				// FIXME need to redo the asymmetric part of the db
			} else {
				List asymmList = asymmetrics.getChildren();
				for (int i = 0; i < asymmList.size(); i++) {
					Element child = (Element) asymmList.get(i);
					if (null == child) {
						continue;
					}
					String algoName = child.getName();
					Set<Integer> keySizes = new HashSet<Integer>();
					List keySizesList = child.getChildren();
					for (int j = 0; j < keySizesList.size(); j++) {
						Element keySize = (Element) keySizesList.get(j);
						if (keySize == null) {
							continue;
						}
						Integer ks = Integer.parseInt(keySize.getText());
						keySizes.add(ks);
					}
					this.asymmetricStrenghts.put(algoName, keySizes);
				}
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			dbValid = false;
		} catch (JDOMException e) {
			e.printStackTrace();
			dbValid = false;
		} catch (IOException e) {
			e.printStackTrace();
			dbValid = false;
		}
		if (!dbValid) {
			this.updateDb();
		}
	}

	/**
	 * Save the algorithm db to a file.
	 * This would be called from the desctructor if java knew one.
	 * So we've got to call it ourselfes somewhere.
	 * @param xmlFile
	 */
	private void saveDb(File xmlFile) {
		// TODO change filename
		//File xmlFile = new File("algodb.xml");
		if (xmlFile.exists() && !xmlFile.canWrite()) {
			throw new IllegalArgumentException("File not accessible! at Location " + xmlFile.getAbsolutePath());
		}
		Element root = new Element(ALGODB_ROOTNODE);
		root.setAttribute("AlgorithmDb", "true");
		root.setAttribute("Version", "1.0");
		// store the provider information
		Element provider = new Element(ALGODB_PROVIDER);
		Provider[] providers = Security.getProviders();
		for (int i = 0; i < providers.length; i++) {
			Provider p = providers[i];
			Element prov = new Element(p.getName());
			prov.setText(String.valueOf(p.hashCode()));
			provider.addContent(prov);
		}
		root.addContent(provider);
		// store the information about the digest algorithms
		Element digest = new Element(ALGODB_DIGEST);
		for (Iterator<String> iterator = this.digestAlgorithm.iterator(); iterator.hasNext();) {
			String d = (String) iterator.next();
			Element e = new Element(d);
			digest.addContent(e);
		}
		root.addContent(digest);
		// store the information about the signature algorithms
		Element signature = new Element(ALGODB_SIGNATURE);
		for (Iterator<String> iterator = this.signatureAlgorithms.iterator(); iterator.hasNext();) {
			String s = (String) iterator.next();
			Element e = new Element(s);
			signature.addContent(e);
		}
		root.addContent(signature);
		// store the information about the symmetric encryption algorithms
		/*// FIMXE disabled because SAFER+ causes a bug and i don't want to fix it. - ds
		Element symmetric = new Element(ALGODB_SYMMETRIC);
		for (Iterator iterator = this.symmetricStrenghts.keySet().iterator(); iterator.hasNext();) {
			String s = (String) iterator.next();
			Set<Integer> strength = this.symmetricStrenghts.get(s);
			Element algorithm = new Element(s);
			for (Iterator iterator2 = strength.iterator(); iterator2.hasNext();) {
				Integer integer = (Integer) iterator2.next();
				Element str = new Element(ALGODB_STRENGTH);
				str.setText(String.valueOf(integer));
				algorithm.addContent(str);
			}
			symmetric.addContent(algorithm);
		}
		root.addContent(symmetric);
		*/
		// store the information about the asymmetric encryption algorithms
		Element asymmetric = new Element(ALGODB_ASYMMETRIC);
		for (Iterator<String> iterator = this.asymmetricStrenghts.keySet().iterator(); iterator.hasNext();) {
			String a = (String) iterator.next();
			Set<Integer> strength = this.asymmetricStrenghts.get(a);
			Element algorithm = new Element(a);
			for (Iterator<Integer> iterator2 = strength.iterator(); iterator2.hasNext();) {
				Integer integer = (Integer) iterator2.next();
				Element str = new Element(ALGODB_STRENGTH);
				str.setText(String.valueOf(integer));
				algorithm.addContent(str);
			}
			asymmetric.addContent(algorithm);
		}
		root.addContent(asymmetric);
		// finalize the xml-document
		Document xml = new Document(root);
		XMLOutputter xmlOut = new XMLOutputter();
		xmlOut.setFormat(Format.getPrettyFormat());
		try {
			xmlOut.output(xml, new FileOutputStream(xmlFile));
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Start the jobs which build the algorithm database.
	 */
	private void updateDb() {
		Job asym = new AsymDetectionJob("Algorithm detection (Asymmetric)");
		asym.setUser(true);
		asym.schedule();
		Job sym = new SymmDectectionJob("Algorithm detection (Symmetric)");
		sym.setUser(true);
		sym.schedule();
		Job digest = new DigestDetectionJob("Algorithm detection (Digest)");
		digest.setUser(true);
		digest.schedule();
		Job sign = new SignatureDetectionJob("Algorithm detection (Signature)");
		sign.setUser(true);
		sign.schedule();
	}

}
