package minica.crypto;

import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.KeyPair;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.List;

import javax.crypto.NoSuchPaddingException;

import minica.util.StringTools;
import codec.Base64;
import codec.asn1.ASN1Exception;
import codec.asn1.DEREncoder;
import codec.pkcs12.PFX;
import codec.x509.X509Certificate;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * Export methods.
 */
public class Export {
	public static final String	CERT_START_TAG	= "-----BEGIN CERTIFICATE-----";

	public static final String	CERT_END_TAG	= "-----END CERTIFICATE-----";

	public static final String	PK_START_TAG	= "-----BEGIN RSA PRIVATE KEY-----";

	public static final String	PK_END_TAG		= "-----END RSA PRIVATE KEY-----";

	/**
	 * use writeCertificateDER instead
	 */
	public static void writeCertificateBase64(Certificate cert, File file) throws CertificateEncodingException,
			IOException, NoSuchAlgorithmException, InvalidKeySpecException, InvalidKeyException,
			NoSuchPaddingException, InvalidAlgorithmParameterException, InterruptedException {
		String certString = Base64.encode(cert.getEncoded());
		certString = StringTools.lineBreakAfterNchars(certString, 65); // break after 65 chars

		BufferedOutputStream bw = new BufferedOutputStream(new FileOutputStream(file));
		bw.write((CERT_START_TAG + "\n").getBytes());
		bw.write(certString.getBytes());
		bw.write(("\n" + CERT_END_TAG).getBytes());
		bw.close();
	}

	public static void writePrivateKeyAndCertificateBase64(Certificate cert, PrivateKey pk, File file) throws CertificateEncodingException,
			IOException, NoSuchAlgorithmException, InvalidKeySpecException, InvalidKeyException,
			NoSuchPaddingException, InvalidAlgorithmParameterException, InterruptedException {
		String certString = Base64.encode(cert.getEncoded());
		String pkString = Base64.encode(pk.getEncoded());
		certString = StringTools.lineBreakAfterNchars(certString, 65); // break after 65 chars
		BufferedOutputStream bw = new BufferedOutputStream(new FileOutputStream(file));
		bw.write((PK_START_TAG + "\n").getBytes());
		bw.write(pkString.getBytes());
		bw.write(("\n" + PK_END_TAG + "\n").getBytes());
		bw.write((CERT_START_TAG + "\n").getBytes());
		bw.write(certString.getBytes());
		bw.write(("\n" + CERT_END_TAG + "\n").getBytes());
		bw.close();
	}

	public static void writeCertificateDER(Certificate cert, File file) throws Exception {
		DEREncoder enc = new DEREncoder(new FileOutputStream(file));
		new X509Certificate(cert.getEncoded()).encode(enc);
		enc.close();
	}

	public static PFX createPKCS12(KeyPair kp, X509Certificate xc, String password)
			throws CertificateEncodingException, GeneralSecurityException, ASN1Exception, FileNotFoundException {
		PFX pfx = new PFX(kp.getPrivate(), xc, null, password.toCharArray(), "FlexiCA", null);
		return pfx;
	}

	public static void writeToPkcs12File(KeyPair kp, X509Certificate xc, String password, String filename)
			throws CertificateEncodingException, IOException, ASN1Exception {
		writeToPkcs12File(kp, xc, password, filename, true, false, false, false, false);
	}

	public static void writeToPkcs12File(KeyPair kp, X509Certificate xc, String password, String filename,
			boolean exportP12, boolean exportDER, boolean exportTXT, boolean exportPEM, boolean exportApache)
			throws IOException, ASN1Exception, CertificateEncodingException {

		PFX pfx;
		try {
			pfx = createPKCS12(kp, xc, password);

			/*
			 * @see http://de.wikipedia.org/wiki/X.509
			 *  Dateinamenserweiterungen für Zertifikate [Bearbeiten]
			 *
			 * Übliche Dateinamenserweiterungen für X.509-Zertifikate sind:
			 *
			 * .CER – DER-kodiertes Zertifikat oder Zertifikatsfolgen
			 * .CRT – DER- oder Base64-kodiertes Zertifikat
			 * .DER – DER-kodiertes Zertifikat
			 * .P12 – PKCS#12, kann öffentliche Zertifikate und private Schlüssel (Kennwort-geschützt) enthalten.
			 * .P7B – Siehe .p7c
			 * .P7C – PKCS#7-signierte Datenstruktur ohne Dateninhalt, nur mit Zertifikat(en) oder Zertifikatsperrliste(n)
			 * .PEM – Base64-kodiertes Zertifikat, umschlossen von “-----BEGIN CERTIFICATE-----" und “-----END CERTIFICATE-----"
			 * .PFX – Siehe .p12
			 *
			 */
			// PFX/PKCS#12 File
			String p12Filename = filename + "." + KeyStoreProvider.PKCS12_EXTENSION;
			// DER-Coded
			String derFilename = filename + ".der";
			// Base64 coded Certificate and Base64 coded private key
			String pemFilename = filename + ".pem";
			// Text Representation
			String txtFilename = filename + ".crt";
			// Apache-Style Configuration File
			String apacheConfFilename = filename + "_apache.conf";
			FileOutputStream output;
			DEREncoder enc;

			if (exportP12) {
				// export as PFX/PKCS#12 File - modified keystore.
				enc = new DEREncoder(new FileOutputStream(p12Filename));
				pfx.encode(enc);
				enc.close();
			}

			if (exportDER) {
				// export as DER - binary encoding of the certificate
				output = new FileOutputStream(derFilename);
				output.write(xc.getEncoded());
				output.close();
			}

			if (exportTXT) {
				// export as TXT - String representation of the certificate
				output = new FileOutputStream(txtFilename);
				output.write((xc.toString()).getBytes());
				output.close();
			}

			// only export the certificate as pem if not exporting apache credntials, which would
			// do this anyway.
			if (exportPEM && !exportApache) {
				// export PEM - Base64 encoding of the binary certificate
				try {
					writeCertificateBase64(xc, new File(pemFilename));
				} catch (InvalidKeyException e) {
					e.printStackTrace();
				} catch (NoSuchAlgorithmException e) {
					e.printStackTrace();
				} catch (InvalidKeySpecException e) {
					e.printStackTrace();
				} catch (NoSuchPaddingException e) {
					e.printStackTrace();
				} catch (InvalidAlgorithmParameterException e) {
					e.printStackTrace();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}

			}

			if (exportApache) {
				File pemFile = new File(pemFilename);
				try {
					writePrivateKeyAndCertificateBase64(xc, kp.getPrivate(), pemFile);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				// export apache configuration
				FileWriter fw = new FileWriter(apacheConfFilename);
				BufferedWriter bw = new BufferedWriter(fw);
				List<String> virtualHost = new ArrayList<String>();
				virtualHost.add("SSLCertificateFile " + pemFile.getName());
				virtualHost.add("SSLCertificateKeyFile " + pemFile.getName());
				virtualHost.add("SSLEngine on");
				for (int i = 0, n = virtualHost.size(); i < n; i++) {
					bw.write(virtualHost.get(i));
					bw.newLine();
				}
				bw.close();
				fw.close();
			}
		} catch (GeneralSecurityException e1) {
			e1.printStackTrace();
		}
	}
}
