package minica.crypto;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CRLException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import minica.Application;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import codec.Base64;
import codec.CorruptedCodeException;
import codec.asn1.ASN1Exception;
import codec.x509.X509Certificate;
import codec.x509.X509Crl;

/*
 * Copyright (c) 2007-2008 Bastian Neuburger Copyright (c) 2007-2008 Dominik Schulz Copyright (c) 2007-2008 Stephan
 * Heuser Copyright (c) 2007-2008 Sören Mess
 * 
 * This file is part of FlexiCA.
 * 
 * FlexiCA is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * FlexiCA is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with FlexiCA; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA </pre>
 */
/**
 * This class contains the certificates and everything related to.
 * @author tex
 *
 */
public class KeyStoreProvider implements ChangeListener {

	/**
	 * The keystore filename.
	 */
	public static final String		KEYSTORE_FILENAME			= "minica.keystore";

	/**
	 * the keystore information db.
	 */
	public static final String		KEYSTORE_INFO_DB_FILENAME	= "minica_infodb.xml";

	public static final String		PKCS12_EXTENSION			= "p12";

	/**
	 * the keystore password.
	 * this is hard coded, since the keystore only contains public information
	 */
	private static final char[]		KEYSTORE_PASSWORD			= { 'a', 'b', 'c' };
	
	//private static final char[]		KEYSTORE_PASSWORD			= Application.getSettings().get("KeyStorePassword").toCharArray(); 
	
	/*
	 * Singleton Pattern
	 */
	private static KeyStoreProvider	instance;

	public synchronized static KeyStoreProvider getInstance() {
		if (null == instance) {
			instance = new KeyStoreProvider();
		}
		return instance;
	}

	/**
	 * the next available serial number.
	 */
	private BigInteger						nextSerialNumber;

	private ArrayList<IModelChangeListener>	listeners	= new ArrayList<IModelChangeListener>();

	private KeyStore						keyStore;

	private List<ICaCertificate>			certificates;

	private KeyStoreProvider() {
		this.nextSerialNumber = BigInteger.ONE;
		this.certificates = new ArrayList<ICaCertificate>();
		try {
			this.keyStore = KeyStore.getInstance("JKS");
			File ksFile = new File(Application.getHome() + KEYSTORE_FILENAME);
			if (ksFile.exists() && ksFile.canRead()) {
				FileInputStream fis = new FileInputStream(ksFile);
				//System.out.println("KeyStoreProvider() - Reading KEystore from: " + ksFile.getAbsolutePath());
				this.keyStore.load(fis, KEYSTORE_PASSWORD);
			} else {
				//System.out.println("KeyStoreProvider() - ks not found. using null.");
				this.keyStore.load(null, KEYSTORE_PASSWORD);
			}
		} catch (KeyStoreException e) {
			e.printStackTrace();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
		} catch (CertificateException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		//System.out.println("KeyStore contructor.");
		/*
		try {
			Enumeration<String> e = this.keyStore.aliases();
			while (e.hasMoreElements()) {
				String str = e.nextElement();
				System.out.println("KeyStore Entry(ks serialize): " + str);
			}
		} catch (KeyStoreException e) {
			System.out.println("KeyStore serialize, KeyStore Exception.");
		}
		*/
		this.deserialize();
	}

	public void addCert(ICaCertificate cert) {
		// handle root-cert
		if (cert != null) {
			if (cert.getParentCa() == null) {
				if (!this.certificates.contains(cert)) {
					this.certificates.add(cert);
					cert.addChangeListener(this);
					this.stateChanged(new ChangeEvent(this));
				}
			}
		}
	}

	public void addChangeListener(IModelChangeListener listener) {
		listeners.add(listener);
	}

	@SuppressWarnings("unchecked")
	public synchronized void deserialize() {
		//System.out.println("KeyStoreProvider.deserialize() - Done.");
		File xmlFile = new File(Application.getHome() + KEYSTORE_INFO_DB_FILENAME);
		// deserializeation of the data
		if (!xmlFile.exists() || !xmlFile.canRead()) {
			//throw new IllegalArgumentException("File not accessible! at Location " + xmlFile.getAbsolutePath());
			return;
		}
		SAXBuilder xmlBuilder = new SAXBuilder();
		Document doc;
		try {
			doc = xmlBuilder.build(new FileInputStream(xmlFile));
			Element root = doc.getRootElement();
			if (root.getAttributeValue("NextSerial") != null)
				this.nextSerialNumber = new BigInteger(root.getAttributeValue("NextSerial"));
			else
				this.nextSerialNumber = BigInteger.ONE;
			//System.out.println("deserialize() - set next Serial to: " + this.nextSerialNumber.toString());
			List rootElements = root.getChildren();
			/*
			 * XML-Root
			 * |- Root-Cert.
			 * |  |- Ca-Cert
			 * |  |  |- Ca-Cert
			 * |  |  |- Client-Cert
			 * |  |- Ca-Cert
			 * |  |  |- Client-Cert
			 * |- Root-Cert
			 *    |- Client-Cert
			 *    |- Client-Cert
			 */
			// parse the data
			for (int i = 0; i < rootElements.size(); i++) {
				Element node = (Element) rootElements.get(i);
				if (null == node) {
					continue;
				}
				//System.out.println("deserialize() - deserializing: " + node.getName() + " " + node.getAttributeValue("alias"));
				this.addCert(this.deserialize(node, null));
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (JDOMException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		this.stateChanged(new ChangeEvent(this));
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final KeyStoreProvider other = (KeyStoreProvider) obj;
		if (certificates == null) {
			if (other.certificates != null)
				return false;
		} else if (!certificates.equals(other.certificates))
			return false;
		return true;
	}

	public List<ICaCertificate> getAllCaCerts() {
		List<ICaCertificate> certs = new ArrayList<ICaCertificate>();
		for (int i = 0; i < this.certificates.size(); i++) {
			certs.add(this.certificates.get(i));
			certs.addAll(this.certificates.get(i).getAllCaCerts());
		}
		return certs;
	}

	public List<BigInteger> getCaCertificateSerials() {
		List<BigInteger> serials = new ArrayList<BigInteger>();
		List<ICaCertificate> certs = getAllCaCerts();
		for (ICaCertificate cert : certs) {
			serials.add(cert.getCertificate().getSerialNumber());
		}
		return serials;
	}

	public ICaCertificate getCertificate(BigInteger serial) {
		List<ICaCertificate> certs = getAllCaCerts();
		for (int i = 0; i < certs.size(); i++) {
			ICaCertificate current = certs.get(i);
			if (current.getCertificate().getSerialNumber().equals(serial))
				return current;
		}
		return null;
	}

	/**
	 * @return the certificates
	 * WARNING: Bad encapsulation!
	 */
	public List<ICaCertificate> getCertificates() {
		return certificates;
	}

	/**
	 * Just return the current max. serial number
	 * w/o incrementing the nextSerial counter.
	 * @return
	 */
	public BigInteger getMaxSerial() {
		return this.nextSerialNumber.subtract(BigInteger.ONE);
	}

	/**
	 * Return the next avail. serial number and
	 * increment the nextSerial counter by one.
	 * @return
	 */
	public BigInteger getNextSerial() {
		BigInteger nextSerial = this.nextSerialNumber;
		this.nextSerialNumber = this.nextSerialNumber.add(BigInteger.ONE);
		return nextSerial;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((certificates == null) ? 0 : certificates.hashCode());
		return result;
	}

	/**
	 * Store all information to the disk.
	 */
	public synchronized void serialize() {
		Element root = new Element("Minica");
		root.setAttribute("Creationtime", String.valueOf(System.currentTimeMillis()));
		root.setAttribute("NextSerial", this.nextSerialNumber.toString());
		for (int i = 0; i < this.certificates.size(); i++) {
			this.serialize(this.certificates.get(i), "Minica", root);
		}
		// store the keystore
		File keyStoreFile = new File(Application.getHome() + KeyStoreProvider.KEYSTORE_FILENAME);
		FileOutputStream fos;
		/*
		FileInputStream fis;
		try {
			fis = new FileInputStream(keyStoreFile);
			this.keyStore.load(fis, KEYSTORE_PASSWORD);
			fis.close();
		} catch (FileNotFoundException e1) {
			try {
				this.keyStore.load(null, KEYSTORE_PASSWORD);
			} catch (NoSuchAlgorithmException e) {
			} catch (CertificateException e) {
			} catch (IOException e) {
			}
		} catch (NoSuchAlgorithmException e) {
			try {
				this.keyStore.load(null, KEYSTORE_PASSWORD);
			} catch (NoSuchAlgorithmException e1) {
			} catch (CertificateException e1) {
			} catch (IOException e1) {
			}
		} catch (CertificateException e) {
			try {
				this.keyStore.load(null, KEYSTORE_PASSWORD);
			} catch (NoSuchAlgorithmException e1) {
			} catch (CertificateException e1) {
			} catch (IOException e1) {
			}
		} catch (IOException e) {
			try {
				this.keyStore.load(null, KEYSTORE_PASSWORD);
			} catch (NoSuchAlgorithmException e1) {
			} catch (CertificateException e1) {
			} catch (IOException e1) {
			}
		}
		*/
		// DEBUG
		/*
		try {
			Enumeration<String> e = this.keyStore.aliases();
			while (e.hasMoreElements()) {
				String str = e.nextElement();
				System.out.println("KeyStore Entry(ks serialize): " + str);
			}
		} catch (KeyStoreException e) {
			System.out.println("KeyStore serialize, KeyStore Exception.");
		}
		*/
		// DEBUG END
		try {
			fos = new FileOutputStream(keyStoreFile);
			this.keyStore.store(fos, KEYSTORE_PASSWORD);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (KeyStoreException e) {
			e.printStackTrace();
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
		} catch (CertificateException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		// store the information db
		// finalize the xml-document
		File xmlFile = new File(Application.getHome() + KEYSTORE_INFO_DB_FILENAME);
		Document xml = new Document(root);
		XMLOutputter xmlOut = new XMLOutputter();
		xmlOut.setFormat(Format.getPrettyFormat());
		try {
			//System.out.println("Writing xml File to " + xmlFile.getAbsolutePath());
			//xmlOut.output(xml, System.out);
			xmlOut.output(xml, new FileOutputStream(xmlFile));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		//System.out.println("KeyStoreProvider.serialize() - Done.");
	}

	public void stateChanged(ChangeEvent e) {
		for (IModelChangeListener l : listeners) {
			l.refresh();
		}
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("KeyStoreProvider[");
		for (int i = 0; i < this.certificates.size(); i++) {
			sb.append(this.certificates.toString());
		}
		sb.append("]");
		return sb.toString();
	}

	/**
	 * Deserialize this node.
	 * @param root
	 * @param parent
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private ICaCertificate deserialize(Element root, ICaCertificate parent) {
		String alias = root.getAttributeValue("alias");
		ICaCertificate ca;
		try {
			java.security.cert.X509Certificate cert2 = (java.security.cert.X509Certificate) this.keyStore
					.getCertificate(alias);
			X509Certificate cert = null;
			byte[] temp;
			try {
				temp = cert2.getEncoded();
				cert = new X509Certificate(temp);
			} catch (CertificateEncodingException e1) {
				e1.printStackTrace();
			}

			if (cert == null) {
				return null;
			}
			//System.out.println("deserialize(p) - cert from keystore: " + cert.getSerialNumber());
			ca = new CaCertificate(cert, parent);
			List nodes = root.getChildren();
			for (int i = 0; i < nodes.size(); i++) {
				Element node = (Element) nodes.get(i);
				if (null == node) {
					continue;
				}
				//System.out.println("deserialize(p) - deserializing: " + node.getName() + " alias: " + node.getAttributeValue("alias"));
				if (node.getName().equals("CaCertificate")) {
					// decode a ca certificate
					// descend recursive ...
					//ca.getCaCerts().add(this.deserialize(node, ca));
					ca.addCert(this.deserialize(node, ca));
				} else if (node.getName().equals("ClientCertificate")) {
					// load a client certificate
					String a = node.getAttributeValue("alias");
					//X509Certificate c = (X509Certificate) this.keyStore.getCertificate(alias);
					// cert2
					java.security.cert.X509Certificate javaCert = (java.security.cert.X509Certificate) this.keyStore
							.getCertificate(a);
					// cert
					X509Certificate codecCert = null;
					// temp
					byte[] myTemp;
					try {
						// temp cert2
						myTemp = javaCert.getEncoded();
						// cert temp
						codecCert = new X509Certificate(myTemp);
					} catch (CertificateEncodingException e1) {
						e1.printStackTrace();
					}
					IClientCertificate clientCert = new ClientCertificate(codecCert, ca);
					ca.addCert(clientCert);
				} else if (node.getName().equals("CRL")) {
					// decode the CRL
					String encodedCRL = node.getText();
					try {
						//System.out.println("deserialize(p) - reading CRL ...");
						byte[] decodedCRL = Base64.decode(encodedCRL);
						ByteArrayInputStream bais = new ByteArrayInputStream(decodedCRL);
						//CertificateFactory cf = CertificateFactory.getInstance("X.509");
						X509Crl crl = new X509Crl(bais);
						//X509CRL crl = (X509CRL) cf.generateCRL(bais);
						bais.close();
						ca.setCRL(crl);
					} catch (CorruptedCodeException e) {
						e.printStackTrace();
					} catch (IOException e) {
						e.printStackTrace();
					} catch (ASN1Exception e) {
						e.printStackTrace();
					}
				} else if (node.getName().equals("dCRL")) {
					// decode the detla-CRL
					String encodedCRL = node.getText();
					try {
						//System.out.println("deserialize(p) - reading CRL ...");
						byte[] decodedCRL = Base64.decode(encodedCRL);
						ByteArrayInputStream bais = new ByteArrayInputStream(decodedCRL);
						//CertificateFactory cf = CertificateFactory.getInstance("X.509");
						X509Crl crl = new X509Crl(bais);
						//X509CRL crl = (X509CRL) cf.generateCRL(bais);
						bais.close();
						ca.setDeltaCrl(crl);
					} catch (CorruptedCodeException e) {
						e.printStackTrace();
					} catch (IOException e) {
						e.printStackTrace();
					} catch (ASN1Exception e) {
						e.printStackTrace();
					}
				}
			}
		} catch (KeyStoreException e) {
			e.printStackTrace();
			// couldn't read the cert from the keystore
			return null;
		}
		return ca;
	}

	private void serialize(ICaCertificate icac, String aliasToken, Element dbNode) {
		if (null == icac || null == icac.getCertificate()) {
			return;
		}
		String myAliasToken = "null";
		Element myDbNode = new Element("CaCertificate");
		myAliasToken = aliasToken + ":" + icac.getCertificate().getSerialNumber();
		myDbNode.setAttribute("serial", String.valueOf(icac.getCertificate().getSerialNumber()));
		myDbNode.setAttribute("alias", myAliasToken);
		// serialize the information and certificate
		try {
			this.keyStore.setCertificateEntry(myAliasToken, icac.getCertificate());
		} catch (KeyStoreException e) {
			e.printStackTrace();
		}
		// serialize the sub ca-certs.
		for (int i = 0; i < icac.getSubCertificates().length; i++) {
			this.serialize(icac.getSubCertificates()[i], myAliasToken, myDbNode);
		}
		// serialize the sub-client-certs.
		for (int i = 0; i < icac.getClientCerts().size(); i++) {
			IClientCertificate client = (IClientCertificate) icac.getClientCerts().get(i);
			if (client == null) {
				continue;
			}
			X509Certificate clientX509 = client.getClientCertificate();
			String clientAliasToken = myAliasToken + ":" + clientX509.getSerialNumber();
			try {
				this.keyStore.setCertificateEntry(clientAliasToken, clientX509);
			} catch (KeyStoreException e) {
				e.printStackTrace();
			}
			Element clientDbNode = new Element("ClientCertificate");
			clientDbNode.setAttribute("serial", String.valueOf(clientX509.getSerialNumber()));
			clientDbNode.setAttribute("alias", clientAliasToken);
			myDbNode.addContent(clientDbNode);
		}
		// serialize the crl, save it base64 encoded in the xml.
		try {
			// this is not optimal
			if (icac.getCRL() != null) {
				//System.out.println("CRL Serializing");
				String encodedCRL = Base64.encode(icac.getCRL().getEncoded());
				Element crl = new Element("CRL");
				crl.setText(encodedCRL);
				myDbNode.addContent(crl);
			}
		} catch (CRLException e) {
			e.printStackTrace();
		}
		// serialize the delta-crls
		try {
			if (icac.getDeltaCrl() != null) {
				String encodedCRL = Base64.encode(icac.getDeltaCrl().getEncoded());
				Element dCrl = new Element("dCRL");
				dCrl.setText(encodedCRL);
				myDbNode.addContent(dCrl);
			}
		} catch (CRLException e) {
			e.printStackTrace();
		}
		// attach the dbnode to the db root
		dbNode.addContent(myDbNode);
	}

}
